"""
Creates images based upon the configuration provided in
``imageconfig.py``.
"""

from swiftpipeline.imageconfig import ImageConfig, Image

import matplotlib.pyplot as plt
import numpy as np

from tqdm import tqdm
from p_tqdm import p_map

from matplotlib.colors import LogNorm, Normalize
from matplotlib.patches import Circle
from mpl_toolkits.axes_grid1 import make_axes_locatable

from unyt import unyt_quantity, unyt_array

from swiftsimio.visualisation.projection import project_pixel_grid
from swiftsimio.visualisation.slice import kernel_gamma
from swiftsimio.visualisation.smoothing_length_generation import (
    generate_smoothing_lengths,
)
from swiftsimio.visualisation.rotation import rotation_matrix_from_vector

from swiftsimio import load, mask, SWIFTDataset
from velociraptor import load as load_catalogue
from swiftpipeline.html import ImageWebpageCreator

from pathlib import Path
from enum import Enum, auto

from typing import Optional, List


def latex_float(f):
    units = f"${f.units.latex_repr}$"

    if f < 1e3 and f > 1e-1:
        return f"${f.value:.3g}$ {units}"

    float_str = "${0:.2g}".format(f.value)
    if "e" in float_str:
        base, exponent = float_str.split("e")
        return r"{0} \times 10^{{{1}}}$ {2}".format(base, int(exponent), units)
    else:
        return float_str + f"$ {units}"


class Projection(Enum):
    DEFAULT = auto()
    FACE_ON = auto()
    EDGE_ON = auto()


class Halo(object):
    """
    Stores some helpful properties of an individual halo.
    """

    mass_200_crit: unyt_quantity
    radius_200_crit: unyt_quantity
    mass_100_kpc_star: unyt_quantity
    radius_100_kpc_star: unyt_quantity
    unique_id: int
    position: unyt_array
    L: unyt_array

    def __init__(
        self,
        mass_200_crit: unyt_quantity,
        radius_200_crit: unyt_quantity,
        mass_100_kpc_star: unyt_quantity,
        radius_100_kpc_star: unyt_quantity,
        unique_id: int,
        position: unyt_array,
        L: unyt_array,
    ):

        self.mass_200_crit = mass_200_crit
        self.radius_200_crit = radius_200_crit
        self.mass_100_kpc_star = mass_100_kpc_star
        self.radius_100_kpc_star = radius_100_kpc_star
        self.unique_id = unique_id
        self.position = unyt_array(position, position[0].units)
        self.L = unyt_array(L, L[0].units)

        return


def haloes_to_visualise(config: ImageConfig, catalogue_path: Path) -> List[Halo]:
    """
    Returns the haloes to visualise from the catalogue.

    Parameters
    ----------

    config: ImageConfig
        Complete image configuration object, containing importantly
        the ``raw_images`` list.

    catalogue_path: Path, str
        Path to the catalogue (``/path/to/halo_0000.properties``).

    Returns
    -------

    haloes: List[Halo]
        Haloes to visualise, with their properties filled out.
    """

    catalogue = load_catalogue(catalogue_path)

    mass_200crit = catalogue.masses.mass_200crit
    a = catalogue.a

    mask = mass_200crit > config.minimum_halo_mass

    if config.centrals_only:
        mask = np.logical_and(mask, catalogue.structure_type.structuretype == 10)

    # Generate some randomness
    rng = np.random.default_rng()

    if config.use_binned_image_selection:
        # Generate a further mask
        max_halo_mass = np.max(mass_200crit).to(config.minimum_halo_mass.units)

        bins = np.arange(
            np.log10(config.minimum_halo_mass),
            np.log10(max_halo_mass) + config.bin_width_in_dex,
            config.bin_width_in_dex,
        )

        digitized = np.digitize(
            x=np.log10(mass_200crit[mask].to(config.minimum_halo_mass.units)), bins=bins
        )

        # Now we ensure that there are at most haloes_to_visualise_per_bin
        # in each of the digitized bins. Haloes are valid in bin 1 and above,
        # we cannot use bin 0 as this is haloes below our minimum mass (which
        # should not be present, see the mask above).

        mass_mask = np.zeros(len(digitized), dtype=bool)

        # This loop can definitely be optimised.
        for bin_id in range(1, len(bins) + 1):
            matches = digitized == bin_id
            number_of_matches = matches.sum()

            # If there's too many, we must sub-sample randomly.
            if number_of_matches > config.haloes_to_visualise_per_bin:
                choices = rng.choice(
                    number_of_matches,
                    size=config.haloes_to_visualise_per_bin,
                    replace=False,
                )

                mass_mask[np.where(matches)[0][choices]] = True

            else:
                mass_mask[matches] = True

        # Modify our original mask with our new changes
        mask[mask] = mass_mask

    # Now build the list of halo objects from valid objects, starting
    # with the most massive.

    haloes = []
    halo_id_order = np.argsort(mass_200crit[mask])[::-1]
    halo_ids_valid = np.arange(len(mass_200crit))[mask][halo_id_order]

    for unique_id in halo_ids_valid:
        haloes.append(
            Halo(
                mass_200_crit=catalogue.masses.mass_200crit[unique_id],
                radius_200_crit=catalogue.radii.r_200crit[unique_id] / a,
                mass_100_kpc_star=catalogue.apertures.mass_star_100_kpc[unique_id],
                radius_100_kpc_star=catalogue.apertures.rhalfmass_star_100_kpc[
                    unique_id
                ]
                / a,
                unique_id=unique_id,
                position=[
                    getattr(catalogue.positions, f"{c}cmbp")[unique_id] / a
                    for c in "xyz"
                ],
                L=[
                    getattr(catalogue.angular_momentum, f"l{c}_star")[unique_id]
                    for c in "xyz"
                ],
            )
        )

    return haloes


def create_scatter(
    snapshot: SWIFTDataset,
    halo: Halo,
    image: Image,
    projection: Projection,
    resolution: int,
) -> unyt_array:
    """
    Creates a projected image for a given image class, and snapshot.

    Parameters
    ----------

    snapshot: SWIFTDataset,
        The opened dataset.

    halo: Halo
        Halo with properties to visualise

    image: Image
        Image class to visualise this time around

    projection: Projection
        Which projection to make

    resolution: int
        Image size along each axis.

    Returns
    -------

    grid: unyt.unyt_array
        Output grid, in the requested units.
    """

    region_given_r = lambda r: [
        halo.position[0] - r,
        halo.position[0] + r,
        halo.position[1] - r,
        halo.position[1] + r,
        halo.position[2] - r,
        halo.position[2] + r,
    ]

    radius = image.get_radius(
        stellar_half_mass=halo.radius_100_kpc_star, r_200_crit=halo.radius_200_crit
    )

    particle_data = getattr(snapshot, image.particle_type, "gas")

    region = region_given_r(radius)

    rotation_center = None
    rotation_matrix = None

    # If the L vector is poorly constrained this will complain,
    # but we don't really care.
    with np.testing.suppress_warnings() as sup:
        sup.filter(RuntimeWarning)
        if projection == Projection.EDGE_ON:
            rotation_center = halo.position.to(particle_data.coordinates.units)
            rotation_matrix = rotation_matrix_from_vector(halo.L.v, "y")
        elif projection == Projection.FACE_ON:
            rotation_center = halo.position.to(particle_data.coordinates.units)
            rotation_matrix = rotation_matrix_from_vector(halo.L.v, "z")

    if hasattr(particle_data, "smoothing_lengths"):
        backend = "fast"
    else:
        backend = "histogram"

    common_attributes = dict(
        data=particle_data,
        boxsize=snapshot.metadata.boxsize,
        resolution=resolution,
        region=region,
        mask=None,
        rotation_matrix=rotation_matrix,
        rotation_center=rotation_center,
        parallel=False,
        backend=backend,
    )

    mass_image = project_pixel_grid(project="masses", **common_attributes)

    if image.visualise == "projected_densities":
        # We're done!
        x_range = region[1] - region[0]
        y_range = region[3] - region[2]
        units = 1.0 / (x_range * y_range)
        units.convert_to_units(1.0 / (x_range.units * y_range.units))
        units *= particle_data.masses.units

        grid = unyt_array(mass_image, units=units)
    else:
        # Need to make the complementary image.
        cache_name = f"_CACHE_MASSWEIGHTED_{image.visualise}"
        cache = getattr(particle_data, cache_name, None)
        particle_array = getattr(particle_data, image.visualise)

        if cache is None:
            cache = particle_array * particle_data.masses

            setattr(
                particle_data,
                cache_name,
                cache,
            )

        weighted_image = project_pixel_grid(project=cache_name, **common_attributes)

        # Deal with zeroes:
        mass_image[mass_image == 0.0] = 1.0

        # K * 1e10 Msun -> K * Msun for the 'units' internally. So we need
        # to reconstruct the true ratio, although this should be ideally the
        # same as particle_array.units.
        units = cache.units / particle_data.masses.units

        grid = unyt_array(weighted_image / mass_image, units=units)

    # Fill if required
    output_units = None
    unit_steal = [image.output_units, image.vmin, image.vmax, grid]

    while output_units is None:
        potential_unit = unit_steal.pop(0)
        if potential_unit is not None:
            output_units = potential_unit.units

    grid.convert_to_units(output_units)

    if image.fill_below is not None:
        mask = grid < image.fill_below.to(output_units)
        grid[mask] = image.fill_below.to(output_units)

    return grid


def save_figure_from_scatter(
    scatter: unyt_array,
    config: ImageConfig,
    halo: Halo,
    image: Image,
    projection: Projection,
    output_path: Path,
):
    """

    Produces the figure associated with a given scatter.

    Parameters
    ----------

    scatter: unyt_array,
        Array of the scattered image.

    config: ImageConfig
        The global image configuration.

    halo: Halo
        Halo with properties to visualise

    image: Image
        Image class to visualise this time around

    projection: Projection
        Which projection to make

    output_path: Path
        Path to save the figure at
    """

    output_filename = (
        output_path
        / f"{image.base_name}_{projection.name.lower()}.{config.image_format}"
    )

    fig, ax = plt.subplots(
        figsize=[config.figure_size] * 2,
        dpi=config.resolution // config.figure_size,
        constrained_layout=False,
        tight_layout=False,
    )

    fig.subplots_adjust(0, 0, 1, 1, 0, 0)
    ax.axis("off")

    vmin = image.vmin.to(scatter.units) if image.vmin is not None else None
    vmax = image.vmax.to(scatter.units) if image.vmax is not None else None

    if image.log_norm:
        norm = LogNorm(vmin=vmin, vmax=vmax, clip=True)
    else:
        norm = Normalize(vmin=vmin, vmax=vmax, clip=True)

    radius = image.get_radius(
        stellar_half_mass=halo.radius_100_kpc_star, r_200_crit=halo.radius_200_crit
    )

    extent_given_r = lambda r: [
        halo.position[0] - r,
        halo.position[0] + r,
        halo.position[1] - r,
        halo.position[1] + r,
    ]

    extent = extent_given_r(radius)

    imshow = ax.imshow(
        scatter.v.T,
        origin="lower",
        norm=norm,
        cmap=image.cmap,
        extent=extent,
    )

    color_bar_ax = fig.add_axes([0.05, 0.95, 0.9, 0.03])

    color_bar = fig.colorbar(
        mappable=imshow,
        ax=ax,
        cax=color_bar_ax,
        orientation="horizontal",
    )

    color_bar_label = image.visualise.title().replace("_", " ")
    color_bar_label = f"{color_bar_label} $\\left[{scatter.units.latex_repr}\\right]$"

    color_bar.set_label(color_bar_label, color=image.text_color)
    color_bar.ax.xaxis.set_tick_params(color=image.text_color)
    color_bar.outline.set_edgecolor(image.text_color)
    plt.setp(plt.getp(color_bar.ax.axes, "xticklabels"), color=image.text_color)

    if image.log_norm:
        color_bar.minorticks_off()

    # Decorate the plot with some additional information

    two_thirds_circle = Circle(
        halo.position[:2],
        radius=(radius * (2.0 / 3.0)).to(halo.position.units).v,
        color=image.text_color,
        linestyle="dashed",
        fill=False,
    )

    ax.add_artist(two_thirds_circle)

    x = halo.position[0]
    y = halo.position[1] - (2.05 / 3.0) * radius
    two_thirds_label = f"{(radius * (2.0 / 3.0)).to('kpc'):3.1f}"

    ax.text(x, y, two_thirds_label, ha="center", va="top", color=image.text_color)

    ax.text(
        0.025,
        0.025,
        (f"Halo {halo.unique_id}\n{projection.name.title().replace('_', ' ')}"),
        ha="left",
        va="bottom",
        transform=ax.transAxes,
        color=image.text_color,
    )

    ax.text(
        0.975,
        0.025,
        (
            f"$M_{{\\rm 200, crit}}$={latex_float(halo.mass_200_crit.to('Solar_Mass'))}\n"
            f"$M_{{*, 100}}$={latex_float(halo.mass_100_kpc_star.to('Solar_Mass'))}"
        ),
        color=image.text_color,
        ha="right",
        va="bottom",
        transform=ax.transAxes,
    )

    fig.savefig(output_filename)

    plt.close(fig)

    return


def save_thumbnail_from_scatter(
    scatter: unyt_array,
    config: ImageConfig,
    halo: Halo,
    image: Image,
    projection: Projection,
    output_path: Path,
):
    """

    Produces the thumbnail image associated with a given scatter.
    Will save at thumbnail.{image type}.

    Parameters
    ----------

    scatter: unyt_array,
        Array of the scattered image.

    config: ImageConfig
        The global image configuration.

    halo: Halo
        Halo with properties to visualise

    image: Image
        Image class to visualise this time around

    projection: Projection
        Which projection to make

    output_path: Path
        Path to save the figure at
    """

    output_filename = output_path / f"thumbnail.{config.image_format}"

    fig, ax = plt.subplots(
        figsize=[1] * 2, dpi=128, constrained_layout=False, tight_layout=False
    )

    fig.subplots_adjust(0, 0, 1, 1, 0, 0)
    ax.axis("off")

    vmin = image.vmin.to(scatter.units) if image.vmin is not None else None
    vmax = image.vmax.to(scatter.units) if image.vmax is not None else None

    if image.log_norm:
        norm = LogNorm(vmin=vmin, vmax=vmax, clip=True)
    else:
        norm = Normalize(vmin=vmin, vmax=vmax, clip=True)

    radius = image.get_radius(
        stellar_half_mass=halo.radius_100_kpc_star, r_200_crit=halo.radius_200_crit
    )

    extent_given_r = lambda r: [
        halo.position[0] - r,
        halo.position[0] + r,
        halo.position[1] - r,
        halo.position[1] + r,
    ]

    extent = extent_given_r(radius)

    ax.imshow(
        scatter.v.T,
        origin="lower",
        norm=norm,
        cmap=image.cmap,
        extent=extent,
    )

    fig.savefig(output_filename)

    plt.close(fig)

    return


def visualise_halo(
    output_path: Path,
    snapshot_path: Path,
    config: ImageConfig,
    halo: Halo,
):
    """
    Creates all of the visualisations in the config for the
    specified halo, and saves them to disk.

    Parameters
    ----------

    output_path: Path, str
        Output path to save images to. Inside this path, there will be
        a number of directories created (one per halo). This path must
        already exist.

    snapshot_path: Path,
        Path to the snapshot. For a sufficiently large volume, and
        a sufficiently small number of haloes, there will be little-to
        -no overlap in the read regions.

    config: ImageConfig
        Opened configuration file.

    halo: Halo
        The halo to read the data for and visualise.
    """

    # First need to find the maximum radius, amongst any
    # of the images.
    radii = [
        image.get_radius(
            stellar_half_mass=halo.radius_100_kpc_star, r_200_crit=halo.radius_200_crit
        )
        for image in config.images
    ]

    max_radius = max(radii)

    extent_given_r = lambda r: [
        [halo.position[x] - r, halo.position[x] + r] for x in range(3)
    ]

    halo_mask = mask(filename=snapshot_path, spatial_only=True)
    halo_mask.constrain_spatial(restrict=extent_given_r(max_radius))

    data = load(filename=snapshot_path, mask=halo_mask)

    # Generate the smoothing lengths if required.
    if config.calculate_dark_matter_smoothing_lengths:
        data.dark_matter.smoothing_lengths = generate_smoothing_lengths(
            coordinates=data.dark_matter.coordinates,
            boxsize=data.metadata.boxsize,
            kernel_gamma=kernel_gamma,
        )

    if config.recalculate_stellar_smoothing_lengths and hasattr(data, "stars"):
        if len(data.stars.coordinates) > 0:
            data.stars.smoothing_lengths = generate_smoothing_lengths(
                coordinates=data.stars.coordinates,
                boxsize=data.metadata.boxsize,
                kernel_gamma=kernel_gamma,
            )

    halo_directory = output_path / f"halo_{halo.unique_id}"
    halo_directory.mkdir(exist_ok=True)

    for image in config.images:
        # Which projections should we make?
        projections = [Projection.DEFAULT]

        if image.face_on:
            projections.append(Projection.FACE_ON)

        if image.edge_on:
            projections.append(Projection.EDGE_ON)

        for projection in projections:
            scatter = create_scatter(
                snapshot=data,
                halo=halo,
                image=image,
                projection=projection,
                resolution=config.resolution,
            )

            save_figure_from_scatter(
                scatter=scatter,
                config=config,
                halo=halo,
                image=image,
                projection=projection,
                output_path=halo_directory,
            )

            if (
                projection == Projection.DEFAULT
                and image.base_name == config.thumbnail_image
            ):
                save_thumbnail_from_scatter(
                    scatter=scatter,
                    config=config,
                    halo=halo,
                    image=image,
                    projection=projection,
                    output_path=halo_directory,
                )


def build_webpage(config: ImageConfig, haloes: List[Halo], output_path: Path):
    """
    Builds and aves the webpages.
    """
    creator = ImageWebpageCreator(
        haloes=haloes,
        config=config,
    )

    creator.save_html(output_path=output_path)

    return


def create_all_images(
    config: ImageConfig,
    output_path: Path,
    snapshot_path: Path,
    catalogue_path: Path,
    parallel: bool = False,
    debug: bool = False,
):
    """
    Create all images, given a config and a set of snapshots
    and catalogues.

    Parameters
    ----------

    config: ImageConfig
        Complete image configuration object, containing importantly
        the ``raw_images`` list.

    output_path: Path, str
        Output path to save images to. Inside this path, there will be
        a number of directories created (one per halo). This path must
        already exist.

    snapshot_path: Path, str
        Path to the snapshot (``/path/to/output_0000.hdf5``).

    catalogue_path: Path, str
        Path to the catalogue (``/path/to/halo_0000.properties``).

    parallel: bool, optional
        Whether or not to create all images in parallel with each other
        (uses p_tqdm).

    debug: bool, optional
        Whether or not to print out the progress of the image creation
    """

    haloes = haloes_to_visualise(config=config, catalogue_path=catalogue_path)

    def packed_vis(halo):
        visualise_halo(
            output_path=output_path,
            snapshot_path=snapshot_path,
            config=config,
            halo=halo,
        )

    if parallel:
        p_map(packed_vis, haloes, disable=not debug)
    else:
        list(map(packed_vis, tqdm(haloes, disable=not debug)))

    build_webpage(config=config, haloes=haloes, output_path=output_path)

    pass


"""
Argument parsing wrapper for the additional scripts.
"""

import argparse as ap
from typing import List, Optional, Dict, Any
from swiftpipeline.config import Config


class ScriptArgumentParser(object):
    """
    Script argument parser for ``swiftpipeline`` additional scripts.

    They must conform to the following command-line api:

    + ``-s``: List of snapshots (not including directory)
    + ``-c``: List of catalogues (not including directory)
    + ``-d``: List of input directories (both the snapshot and catalogue
              should be in there).
    + ``-n``: List of run names, optional, for use in legends.
    + ``-o``: Output directory for the figure.
    + ``-C``: Configuration directory (contains the ``config``.yml)
    + ``-a``: Additional args for a given script
    """

    parser: ap.ArgumentParser
    description: str

    # List of snapshots that are to be processed.
    snapshot_list: List[str]
    # List of catalogues (in the same order as snapshots) to be processed.
    catalogue_list: List[str]
    # List of directories that contain the above snapshots and catalogues
    directory_list: List[str]
    # List of representative names for the snapshots; may be a list of Nones
    name_list: List[Optional[str]]
    # Directory to output the figure to
    output_directory: str
    # Configuration directory containing config.yml.
    config_directory: str

    # Number of inputs to the script.
    number_of_inputs: int

    # Config object containing all relevant information.
    config: Config

    # Additional arguments, with the values being the defaults.
    additional_arguments: Dict[str, Any]

    def __init__(self, description: str, additional_arguments: Dict[str, Any] = None):
        """
        Initialises the argument parser object and parses the args, as they say.

        Parameters
        ----------

        description: str
            Description (for the command line) of this script. Shown when the
            script is initialised with -h.

        additional_arguments: dict
            Dictionary containing any additional arguments. Has the structure
            of the key being the command-line name (also used in the config.yml
            file), and the value being the default. For example:

            .. code-block:: python

                additional_arguments = {
                    "name": "No Name",
                    "plot_total": 8,
                }
        """

        self.description = description

        self.additional_arguments = (
            additional_arguments if additional_arguments is not None else dict()
        )

        self.__setup_parser()
        self.__parse_arguments()

        return

    def __setup_parser(self):
        """
        Set up the argument parser.
        """

        self.parser = ap.ArgumentParser(description=self.description)

        self.parser.add_argument(
            "-s",
            "--snapshots",
            help="Snapshot list. Do not include directory. Example: snapshot_0000.hdf5",
            type=str,
            required=True,
            nargs="*",
        )

        self.parser.add_argument(
            "-c",
            "--catalogues",
            help=(
                "Catalogue list. Do not include directory. Example: "
                "catalogue_0000.properties"
            ),
            type=str,
            required=True,
            nargs="*",
        )

        self.parser.add_argument(
            "-d",
            "--input-directories",
            help="Input directory list. Catalogue and snapshot are in this directory.",
            type=str,
            required=True,
            nargs="*",
        )

        self.parser.add_argument(
            "-n",
            "--run-names",
            help="Names of the runs for placement in legends.",
            type=str,
            required=False,
            nargs="*",
        )

        self.parser.add_argument(
            "-o",
            "--output-directory",
            help="Output directory for the produced figure.",
            type=str,
            required=True,
        )

        self.parser.add_argument(
            "-C",
            "--config",
            help="Config directory that contains config.yaml",
            type=str,
            required=True,
        )

        self.parser.add_argument(
            "-a",
            "--additional-args",
            help="Additional command line args for a given script",
            type=str,
            required=False,
            nargs="*",
        )

        for key, value in self.additional_arguments.items():
            self.parser.add_argument(
                f"--{key}",
                help=f"Additional argument {key}. Default: {value}.",
                default=value,
                required=False,
            )

        return

    def __parse_arguments(self):
        """
        Parses the arguments from the ``parser``.
        """

        args = self.parser.parse_args()

        self.number_of_inputs = len(args.snapshots)

        self.snapshot_list = args.snapshots
        self.catalogue_list = args.catalogues
        self.directory_list = args.input_directories
        self.name_list = (
            args.run_names
            if args.run_names is not None
            else [None] * self.number_of_inputs
        )
        self.output_directory = args.output_directory
        self.config_directory = args.config

        # Parse any user-added arguments
        for key, value in self.additional_arguments.items():
            setattr(self, key, getattr(args, key, None))

        self.config = Config(config_directory=self.config_directory)

        return

    @property
    def stylesheet_location(self):
        return f"{self.config_directory}/{self.config.matplotlib_stylesheet}"



"""
Configuration object for the entire pipeline.
"""

import yaml
from typing import List, Union

# Items to read directly from the yaml file with their defaults
direct_read = {
    "auto_plotter_directory": None,
    "auto_plotter_registration": None,
    "auto_plotter_global_mask": None,
    "observational_data_directory": None,
    "matplotlib_stylesheet": "default",
    "description_template": None,
    "custom_css": None,
}


class Script(object):
    """
    Object describing the core properties of a 'script'.
    """

    # Filename of the python script to be ran
    filename: str
    # Output file name(s)
    # List of strings if the script produces several plots; otherwise, string
    # If list, required to have the same size as caption and title
    output_file: Union[str, List[str]]
    # Caption(s) for displaying in a webpage, etc.
    # List of strings if the script produces several plots (one caption per plot); otherwise, string
    caption: Union[str, List[str]]
    # Section heading; used to classify this with similar figures
    # in the output.
    section: str
    # Plot title(s); written above the caption
    # List of strings if the script makes several plots (one title per plot); otherwise, string
    title: Union[str, List[str]]
    # Show on webpage; Defaults to True but used to disable webpage plotting
    # in the config file if required.
    show_on_webpage: bool
    # additional arguments to be fed to a given script
    additional_arguments: dict
    # Use in the case where we have comparisons? The scripts may be disabled
    # in comparison cases for performance reasons.
    use_for_comparison: bool

    def __init__(self, script_dict: dict):
        """
        Takes the dictionary and extracts it to inner variables.
        """

        self.filename = script_dict.get("filename", "")
        self.caption = script_dict.get("caption", "")
        self.output_file = script_dict.get("output_file", "")
        self.section = script_dict.get("section", "")
        self.title = script_dict.get("title", "")
        self.show_on_webpage = script_dict.get("show_on_webpage", True)
        self.additional_arguments = script_dict.get("additional_arguments", {})
        self.use_for_comparison = script_dict.get("use_for_comparison", True)
        return

    def __str__(self):
        return (
            f"Script: {self.filename}, produces {self.filename} in {self.section} "
            f"described as {self.caption}"
        )

    def __repr__(self):
        return f"Script object describing the {self.filename} script."

    @property
    def additional_argument_list(self):
        """
        Gets a list of additional arguments, with --key, value ordering.
        """

        additional_arguments = []

        for key, value in self.additional_arguments.items():
            additional_arguments.append(f"--{key}")
            additional_arguments.append(f"{value}")

        return additional_arguments


class Config(object):
    """
    Configuration object containing the major parameters read from the
    yaml file.
    """

    # Raw config read directly from the file, before processing.
    raw_config: dict
    raw_scripts: List[Script]

    # Set up the object.
    __slots__ = list(direct_read.keys()) + [
        "raw_scripts",
        "config_directory",
        "raw_config",
    ]

    def __init__(self, config_directory: str):
        """
        Set up and load the config to this object.

        Parameters
        ----------

        config_directory: str
            Directory containing the configuration ``config.yml``.
        """

        self.config_directory = config_directory

        self.__read_config()
        self.__extract_to_variables()
        self.__extract_scripts()

        return

    def __str__(self):
        return f"Configuration file object describing {self.config_directory}"

    def __repr__(self):
        return self.__str__()

    def __read_config(self):
        """
        Read the ``config.yml`` and store the basic contents in
        ``self.raw_config``.
        """

        with open(f"{self.config_directory}/config.yml", "r") as handle:
            self.raw_config = yaml.safe_load(handle)

        return

    def __extract_to_variables(self):
        """
        Extracts items from the configuration dictionary to local variables.
        """

        for variable, default in direct_read.items():
            setattr(self, variable, self.raw_config.get(variable, default))

        return

    def __extract_scripts(self):
        """
        Extracts the items in the scripts section to their own
        list with objects describing each script.
        """

        raw_scripts = self.raw_config.get("scripts", [])
        self.raw_scripts = [
            Script(script_dict=script_dict) for script_dict in raw_scripts
        ]

        return

    @property
    def scripts(self):
        """
        Gets all of the scripts defined in the parameter file.
        """
        return self.raw_scripts

    @property
    def comparison_scripts(self):
        """
        Gets the scripts only to be used in comparisons from the parameter
        file.
        """
        return [script for script in self.raw_scripts if script.use_for_comparison]



"""
Functions that aid in the production of the HTML webpages.
"""

from swiftpipeline import __version__ as pipeline_version
from swiftpipeline.config import Config

from swiftsimio import SWIFTDataset

from velociraptor import __version__ as velociraptor_version
from velociraptor.autoplotter.metadata import AutoPlotterMetadata

from jinja2 import Environment, PackageLoader, FileSystemLoader, select_autoescape
from time import strftime
from typing import List, Dict
from pathlib import Path

import unyt


def format_number(number):
    """
    Formats a number from float (with or without units) to a latex-like number.
    """

    try:
        units = f"\\; {number.units.latex_repr}"
    except:
        units = ""

    try:
        mantissa, exponent = ("%.3g" % number).split("e+")
        exponent = f" \\times 10^{{{int(exponent)}}}"
    except:
        mantissa = "%.3g" % number
        exponent = ""

    return f"\\({mantissa}{exponent}{units}\\)"


def get_if_present_float(dictionary, value: str, input_unit=None, output_unit=None):
    """
    A replacement for .get() that also formats the number if present.

    Assumes data should be a float.
    """

    try:
        value = float(dictionary[value])

        if input_unit is not None:
            value = unyt.unyt_quantity(value, input_unit)

            if output_unit is not None:
                value.convert_to_units(output_unit)

        return format_number(value)
    except KeyError:
        return ""


def get_if_present_int(dictionary, value: str, input_unit=None, output_unit=None):
    """
    A replacement for .get() that also formats the number if present.

    Assumes data should be an integer.
    """

    try:
        value = int(dictionary[value])

        if input_unit is not None:
            value = unyt.unyt_quantity(value, input_unit)

            if output_unit is not None:
                value.convert_to_units(output_unit)

        return format_number(value)
    except KeyError:
        return ""


def camel_to_title(string):
    return string.title().replace("_", " ")


class WebpageCreator(object):
    """
    Creates webpages based on the information that is provided in
    the plots metadata through the autoplotter and the additional
    plotting interface provided through the pipeline.
    """

    environment: Environment
    loader: PackageLoader

    variables: dict
    html: str

    auto_plotter_metadata: AutoPlotterMetadata
    config: Config

    def __init__(self):
        """
        Sets up the ``jinja`` templating system.
        """

        self.loader = PackageLoader("swiftpipeline", "templates")
        self.environment = Environment(
            loader=self.loader, autoescape=select_autoescape(["js"])
        )

        # Initialise empty variables dictionary, with the versions of
        # this package and the velociraptor package used.
        self.variables = dict(
            pipeline_version=pipeline_version,
            velociraptor_version=velociraptor_version,
            creation_date=strftime(r"%Y-%m-%d"),
            sections={},
            runs=[],
        )

        return

    def render_webpage(self, template: str = "plot_viewer.html") -> str:
        """
        Renders a webpage based on the internal variables stored in
        the ``variables`` dictionary.

        Parameters
        ----------

        template: str
            The name of the template that you wish to use. Defaults to
            "plot_viewer.html".

        Returns
        -------

        html: str
            The resulting HTML. This is also stored in ``.html``.
        """

        self.html = self.environment.get_template(template, parent="base.html").render(
            **self.variables
        )

        return self.html

    def add_metadata(self, page_name: str):
        """
        Add additional metadata to the page.

        Parameters
        ----------

        page_name: str
            Name to put in the page title.
        """

        self.variables.update(dict(page_name=page_name))

    def add_auto_plotter_metadata(self, auto_plotter_metadata: AutoPlotterMetadata):
        """
        Adds the auto plotter metadata to the section / plot metadata.

        Parameters
        ----------

        auto_plotter_metadata: AutoPlotterMetadata
            The complete instance of ``AutoPlotterMetadata`` after running the
            ``velociraptor`` ``AutoPlotter``.
        """

        self.auto_plotter_metadata = auto_plotter_metadata

        # Unique sections
        sections = {
            plot.section
            for plot in auto_plotter_metadata.metadata
            if plot.show_on_webpage
        }

        for section in sections:
            plots = [
                dict(
                    filename=f"{plot.filename}.{auto_plotter_metadata.file_extension}",
                    title=plot.title,
                    caption=plot.caption,
                    hash=abs(hash(plot.caption + plot.title)),
                )
                for plot in auto_plotter_metadata.metadata
                if plot.section == section and plot.show_on_webpage
            ]

            current_section_plots = (
                self.variables["sections"].get(section, {"plots": []}).get("plots", [])
            )

            self.variables["sections"][section] = dict(
                title=section,
                plots=plots + current_section_plots,
                id=abs(hash(section)),
            )

        return

    def add_config_metadata(self, config: Config, is_comparison: bool = False):
        """
        Adds the section metadata from the additional plots
        defined under ``config.yml::scripts``.

        Parameters
        ----------

        config: Config
            Configuration object from ``swift-pipeline``.

        is_comparison: bool, optional
            Is this webpage being created for a comparison? Default: False.
        """

        self.config = config

        # Unique sections
        sections = {
            script.section for script in config.scripts if script.show_on_webpage
        }

        scripts_to_use = config.comparison_scripts if is_comparison else config.scripts

        for section in sections:

            plots: List[Dict] = []

            for script in scripts_to_use:
                if script.section == section and script.show_on_webpage:

                    # Check whether we expect more than one plot (output file) produced by the script
                    if isinstance(script.output_file, list):

                        # If so, check that each plot has its own title and caption
                        assert isinstance(script.title, list) and isinstance(
                            script.caption, list
                        ), (
                            f"Check the config parameters for '{script.filename}'. "
                            f"If 'output_file' is a list object, then 'title' and 'caption' must be too!"
                        )

                        # Check that the number of plots is the same as the number of their titles and captions
                        assert (
                            len(script.output_file)
                            == len(script.title)
                            == len(script.caption)
                        ), (
                            f"Check the config parameters for '{script.filename}'. "
                            f"Lists 'output_file', 'title' and 'caption' must have the same size!"
                        )

                        # Loop over plots made by the script
                        for output_file, title, caption in zip(
                            script.output_file, script.title, script.caption
                        ):

                            # Save everything into a dict
                            plot = dict(
                                filename=output_file,
                                title=title,
                                caption=caption,
                                hash=abs(hash(caption + output_file)),
                            )

                            # Add collect in a list
                            plots.append(plot)

                    # The script makes just a single plot
                    else:
                        plot = dict(
                            filename=script.output_file,
                            title=script.title,
                            caption=script.caption,
                            hash=abs(hash(script.caption + script.output_file)),
                        )
                        plots.append(plot)

            current_section_plots = (
                self.variables["sections"].get(section, {"plots": []}).get("plots", [])
            )

            self.variables["sections"][section] = dict(
                title=section,
                plots=plots + current_section_plots,
                id=abs(hash(section)),
            )

        return

    def add_run_metadata(self, config: Config, snapshots: List[SWIFTDataset]):
        """
        Adds the "run" metadata (using the user-defined description.html).

        Parameters
        ----------

        config: Config
            Configuration object from ``swift-pipeline``.

        snapshots: List[SWIFTDataset].
            SWIFT Datasets used to generate the HTML with.
        """

        loader = FileSystemLoader(config.config_directory)
        environment = Environment(loader=loader)

        environment.filters["format_number"] = format_number
        environment.filters["camel_to_title"] = camel_to_title
        environment.filters["get_if_present_float"] = get_if_present_float
        environment.filters["get_if_present_int"] = get_if_present_int

        if config.description_template is not None:
            self.variables["runs"] = [
                dict(
                    description=environment.get_template(
                        config.description_template
                    ).render(data=data),
                )
                for data in snapshots
            ]

        if config.custom_css is not None:
            self.variables["custom_css"] = environment.get_template(
                config.custom_css
            ).render()

        return

    def save_html(self, filename: str):
        """
        Saves the html in ``self.html`` to the filename provided.

        Parameters
        ----------

        filename: str
            Full filename (including file path) to save the HTML as.
        """

        with open(filename, "w") as handle:
            handle.write(self.html)


class ImageWebpageCreator(object):
    """
    Creates the webpages for the imaging system.
    """

    environment: Environment
    loader: PackageLoader

    variables: dict
    html: str

    config: Config

    def __init__(self, haloes, config):
        """
        Sets up the ``jinja`` templating system.
        """

        self.loader = PackageLoader("swiftpipeline", "templates")
        self.environment = Environment(
            loader=self.loader, autoescape=select_autoescape(["js"])
        )

        self.environment.filters["format_number"] = format_number
        self.environment.filters["camel_to_title"] = camel_to_title
        self.environment.filters["get_if_present_float"] = get_if_present_float
        self.environment.filters["get_if_present_int"] = get_if_present_int

        # Initialise empty variables dictionary, with the versions of
        # this package and the velociraptor package used.
        self.variables = dict(
            pipeline_version=pipeline_version,
            velociraptor_version=velociraptor_version,
            creation_date=strftime(r"%Y-%m-%d"),
            haloes=haloes,
            config=config,
            images=config.images,
        )

        return

    def render_gallery(self) -> str:
        """
        Renders a webpage based on the internal variables stored in
        the ``variables`` dictionary.

        Parameters
        ----------

        template: str
            The name of the template that you wish to use. Defaults to
            "plot_viewer.html".

        Returns
        -------

        html: str
            The resulting HTML. This is also stored in ``.html``.
        """

        self.html = self.environment.get_template(
            "image_gallery.html", parent="base.html"
        ).render(page_name="Image Gallery", **self.variables)

        return self.html

    def render_single_halo(self, halo) -> str:
        return self.environment.get_template(
            "image_halo.html", parent="base.html"
        ).render(page_name=f"Halo {halo.unique_id}", halo=halo, **self.variables)

    def save_html(self, output_path: Path):
        """
        Saves all the relevant HTML, to the output path.
        """

        gallery_html = self.render_gallery()
        gallery_filename = Path(output_path) / "index.html"

        with open(gallery_filename, "w") as handle:
            handle.write(gallery_html)

        for halo in self.variables["haloes"]:
            halo_html = self.render_single_halo(halo)
            halo_filename = Path(output_path) / f"halo_{halo.unique_id}/index.html"

            with open(halo_filename, "w") as handle:
                handle.write(halo_html)



import setuptools
from swiftpipeline import __version__

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="swiftpipeline",
    version=__version__,
    description="Pipeline for producing galaxy scaling relation plots.",
    url="https://github.com/SWIFTSIM/pipeline",
    author="Josh Borrow",
    author_email="josh@joshborrow.com",
    packages=setuptools.find_packages(),
    long_description=long_description,
    long_description_content_type="text/markdown",
    zip_safe=False,
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU Lesser General Public License v3 or later (LGPLv3+)",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "astropy",
        "swiftsimio",
        "matplotlib",
        "jinja2",
        "velociraptor",
        "unyt",
        "tqdm",
        "p_tqdm",
    ],
    scripts=["swift-pipeline", "swift-image"],
    include_package_data=True,
)



"""
Configuration for imaging creator.
"""

import yaml
from unyt import unyt_quantity
from typing import List, Optional

# Items to read directly from the yaml file with their defaults
direct_read = {
    "resolution": 512,
    "figure_size": 8,
    "image_format": "jpg",
    "recalculate_stellar_smoothing_lengths": True,
    "calculate_dark_matter_smoothing_lengths": True,
    "centrals_only": True,
    "minimum_halo_mass": unyt_quantity(1e9, "Solar_Mass"),
    "use_binned_image_selection": False,
    "bin_width_in_dex": 0.5,
    "haloes_to_visualise_per_bin": 10,
    "thumbnail_image": "",
}


class Image(object):
    """
    Object describing the properties of a given image, implementing
    defaults.
    """

    # Name (key in the dictionary); will form part of the filename
    base_name: str
    # Name of the group of figures
    name: str
    # Radius out to which to visualise
    radius_raw: List[str]
    radius: unyt_quantity
    # Particle type to visualise
    particle_type: str
    # Quantity to visualise; projected_densities is a special one because
    # it does not lead to dividing out by density.
    visualise: str
    # Vmin, vmax for the color map
    vmin: Optional[unyt_quantity]
    vmax: Optional[unyt_quantity]
    # Values in the output grid below this value will be filled with it
    fill_below: Optional[unyt_quantity]
    # Output units to use on the color bar. If not present, they use the
    # ones in `vmin`, which if not present uses the ones in `vmax`. If no
    # units are present anywhere, we use the internal units.
    output_units: Optional[unyt_quantity]
    # Color of the text overlaid on the image
    text_color: str
    # Background color for the figure
    plot_background_color: str
    # Color map to use
    cmap: str
    # Produce a 'face on' version of this image?
    face_on: bool
    # Produce an 'edge on' version of this image?
    edge_on: bool
    # Produce the default projection? Always on.
    default: bool = True
    # Logarithmically normalise the image?
    log_norm: bool

    def __init__(self, base_name: str, image_dict: dict):
        """
        Extracts the dictionary to internal variables
        """

        self.base_name = base_name
        self.name = image_dict.get("name", "")
        self.radius_raw = image_dict.get("radius", None)
        self.particle_type = image_dict.get("particle_type", "gas")
        self.visualise = image_dict.get("visualise", "projected_densities")
        self.vmin, self.vmax = self._unpack_vminmax(image_dict)
        self.fill_below = self._unpack_fill_below(image_dict)
        self.output_units = self._unpack_output_units(image_dict)
        self.cmap = image_dict.get("cmap", "viridis")
        self.text_color = image_dict.get("text_color", "white")
        self.plot_background_color = image_dict.get("plot_background_color", "white")
        self.face_on = bool(image_dict.get("face_on", True))
        self.edge_on = bool(image_dict.get("edge_on", True))
        self.log_norm = bool(image_dict.get("log_norm", True))

    def _unyt_from_dict_item(self, name, image_dict):
        """
        Extracts an unyt_quantity from items.
        """

        raw_possible = image_dict.get(name, None)

        if raw_possible is not None:
            try:
                size, unit = raw_possible
            except ValueError:
                size = 1.0
                unit = raw_possible

            raw_possible = unyt_quantity(float(size), unit)

        return raw_possible

    def _unpack_vminmax(self, image_dict):
        vmin = self._unyt_from_dict_item(name="vmin", image_dict=image_dict)
        vmax = self._unyt_from_dict_item(name="vmax", image_dict=image_dict)
        return vmin, vmax

    def _unpack_fill_below(self, image_dict):
        return self._unyt_from_dict_item(name="fill_below", image_dict=image_dict)

    def _unpack_output_units(self, image_dict):
        return self._unyt_from_dict_item(name="output_units", image_dict=image_dict)

    def get_radius(
        self, stellar_half_mass: unyt_quantity, r_200_crit: unyt_quantity
    ) -> unyt_quantity:
        """
        Gets the radius from the 'raw' radius. This requires using the half
        mass and r_200_crit values from the given halo, as we use an
        adaptive scale.

        Parameters
        ----------

        stellar_half_mass: unyt_quantity
            Half-mass radius (stars) for the object.

        r_200_crit: unyt_quantity
            Radius at which the density drops to 200 times the critical
            density for this halo.

        Returns
        -------

        vis_radius: unyt_quantity
            Radius to be used for visualisation, as specified in
            the parameter file.
        """

        if self.radius_raw is None:
            raise AttributeError(f"You must specify a radius for image {name}.")

        value, units = self.radius_raw

        if units == "stellar_half_mass":
            units = stellar_half_mass

            if units == 0.0:
                # Big mistake - fall back on r_200_crit
                units = r_200_crit

        if units == "r_200_crit":
            units = r_200_crit

        return unyt_quantity(float(value), units)


class ImageConfig(object):
    """
    Configuration object containing the image config parameters.
    """

    # Raw config read directly from the file, before processing.
    raw_config: dict
    images: List[Image]

    # Set up the object.
    __slots__ = list(direct_read.keys()) + [
        "images",
        "thumbnail_image",
        "raw_config",
        "config_directory",
    ]

    def __init__(self, config_directory: str):
        """
        Set up and load the config to this object.

        Parameters
        ----------

        config_directory: str
            Directory containing the configuration ``config.yml``.
        """

        self.config_directory = config_directory

        self.__read_config()
        self.__extract_to_variables()
        self.__extract_images()

        return

    def __str__(self):
        return f"Image configuration file object describing {self.config_directory}"

    def __repr__(self):
        return self.__str__()

    def __read_config(self):
        """
        Read the ``config.yml`` and store the basic contents in
        ``self.raw_config``.
        """

        with open(f"{self.config_directory}/images.yml", "r") as handle:
            self.raw_config = yaml.safe_load(handle)

        return

    def __extract_to_variables(self):
        """
        Extracts items from the configuration dictionary to local variables.
        """

        for variable, default in direct_read.items():
            try:
                setattr(
                    self,
                    variable,
                    type(default)(self.raw_config.get(variable, default)),
                )
            except RuntimeError:
                value, unit = self.raw_config.get(variable, default)
                setattr(
                    self,
                    variable,
                    unyt_quantity(float(value), unit),
                )

        return

    def __extract_images(self):
        """
        Extracts all of the individual images and their configuration.
        """

        raw_images = self.raw_config.get("images", {})
        self.images = [
            Image(base_name=name, image_dict=image_dict)
            for name, image_dict in raw_images.items()
        ]

        return
